
/** Copyright (C) 2005 10LE
  * GTK Mozilla
  * @file gmozillaglobalhistory.cpp:
  */

#include "gmozillaglobalhistory.h"

#include <nsIBrowserHistory.h>
#include <nsIGlobalHistory2.h>
#include <mdb.h>
#include <nsIMdbFactoryFactory.h>
#include <nsMorkCID.h>
#include <nsILocalFile.h>
#include <nsAppDirectoryServiceDefs.h>
#include <plstr.h>
#include <prprf.h>
#include <nsNetUtil.h>

#define G_MOZILLA_TEMP_HISTORY_FILE_NAME ".history.dat~"

gchar *G_MOZILLA_TEMP_HISTORY_FILE_PATH;

mdb_scope  kToken_HistoryRowScope;
mdb_kind   kToken_HistoryKind;

mdb_column kToken_URLColumn;
mdb_column kToken_ReferrerColumn;
mdb_column kToken_LastVisitDateColumn;
mdb_column kToken_FirstVisitDateColumn;
mdb_column kToken_VisitCountColumn;
mdb_column kToken_NameColumn;
mdb_column kToken_HostnameColumn;
mdb_column kToken_HiddenColumn;
mdb_column kToken_TypedColumn;
mdb_column kToken_GeckoFlagsColumn;

// meta-data tokens
mdb_column kToken_LastPageVisited;
mdb_column kToken_ByteOrder;

nsIMdbEnv*   mEnv;
nsIMdbStore* mStore;
nsIMdbTable* mTable;
nsCOMPtr<nsIMdbRow> mMetaRow;

PRBool mNowValid = PR_FALSE;
PRBool mReverseByteOrder;

nsresult OpenDB(void);
nsresult copyTempFile();

void SwapBytes(const PRUnichar *source, PRUnichar *dest,
                           PRInt32 aLen)
{
    PRUint16 c;
    const PRUnichar *inp;
    PRUnichar *outp;
    PRInt32 i;

    inp = source;
    outp = dest;
    for (i = 0; i < aLen; i++)
    {
        c = *inp++;
        *outp++ = (((c >> 8) & 0xff) | (c << 8));
    }
    return;
}

nsresult GetRowValue(nsIMdbRow *aRow, mdb_column aCol,
                             PRTime *aResult)
{
    mdb_err err;
    mdbYarn yarn;
    
    err = aRow->AliasCellYarn(mEnv, aCol, &yarn);
    
    if (err != 0) return NS_ERROR_FAILURE;

    *aResult = LL_ZERO;
    
    if (!yarn.mYarn_Fill || !yarn.mYarn_Buf) return NS_OK;
    
    PR_sscanf((const char*)yarn.mYarn_Buf, "%lld", aResult);
    
    return NS_OK;
}

nsresult GetRowValue(nsIMdbRow *aRow, mdb_column aCol,
                             nsACString& aResult)
{
    mdb_err err;
    mdbYarn yarn;
    
    err = aRow->AliasCellYarn(mEnv, aCol, &yarn);
    
    if (err != 0) return NS_ERROR_FAILURE;
    
    const char* startPtr = (const char*)yarn.mYarn_Buf;
    
    if (startPtr)
        aResult.Assign(Substring(startPtr, startPtr + yarn.mYarn_Fill));
    else
        aResult.Truncate();
    return NS_OK;
}

nsresult GetRowValue(nsIMdbRow *aRow, mdb_column aCol,
                             nsAString& aResult)
{
    mdb_err err;
    mdbYarn yarn;
  
    err = aRow->AliasCellYarn(mEnv, aCol, &yarn);
  
    if (err != 0) return NS_ERROR_FAILURE;

    aResult.Truncate(0);
    if (!yarn.mYarn_Fill)
        return NS_OK;
  
    switch (yarn.mYarn_Form)
    {
    case 0:                       // unicode
        if (mReverseByteOrder)
        {
            // The file is other-endian; we must byte-swap the result.
            PRUnichar *swapval;
            int len = yarn.mYarn_Fill / sizeof(PRUnichar);
            swapval = (PRUnichar *)g_try_malloc(yarn.mYarn_Fill);
            if (!swapval)
                return NS_ERROR_OUT_OF_MEMORY;
            SwapBytes((const PRUnichar *)yarn.mYarn_Buf, swapval, len);
            aResult.Assign(swapval, len);
            g_free(swapval);
        }
        else
            aResult.Assign((const PRUnichar *)yarn.mYarn_Buf, yarn.mYarn_Fill/sizeof(PRUnichar));
        break;
    // eventually we'll be supporting this in SetRowValue()
    case 1:                       // UTF8
        CopyUTF8toUTF16(Substring((const char*)yarn.mYarn_Buf,
                              (const char*)yarn.mYarn_Buf + yarn.mYarn_Fill),
                                aResult);
        break;
    default:
        return NS_ERROR_UNEXPECTED;
    }
    return NS_OK;
}

nsresult SetRowValue(nsIMdbRow *aRow, mdb_column aCol,
                             const char* aValue)
{
    mdb_err err;
    PRInt32 len = PL_strlen(aValue);
    mdbYarn yarn = { (void*) aValue, len, len, 0, 0, nsnull };
    err = aRow->AddColumn(mEnv, aCol, &yarn);
    if (err != 0) return NS_ERROR_FAILURE;

    return NS_OK;
}

/* Check if a url is a relevant url */
PRBool IsValidEntry(const char* url)
{
    nsresult rv;

    nsCOMPtr<nsIURI> aURI;
    nsCOMPtr<nsIIOService> ios = do_GetIOService();
    if (ios)
    {		
        ios->NewURI(nsDependentCString(url), nsnull, nsnull,
                    getter_AddRefs(aURI));
    }
	
    PRBool isHTTP, isHTTPS;

    rv = aURI->SchemeIs("http", &isHTTP);
    rv |= aURI->SchemeIs("https", &isHTTPS);
    NS_ENSURE_SUCCESS (rv, NS_ERROR_FAILURE);

    if (!isHTTP && !isHTTPS)
    {
        PRBool isAbout, isViewSource, isChrome, isData, isJavascript;
        rv  = aURI->SchemeIs("about", &isAbout);
        rv |= aURI->SchemeIs("view-source", &isViewSource);
        rv |= aURI->SchemeIs("chrome", &isChrome);
        rv |= aURI->SchemeIs("data", &isData);
        rv |= aURI->SchemeIs("javascript", &isJavascript);

        NS_ENSURE_SUCCESS(rv, NS_ERROR_FAILURE);

        if (isAbout || isViewSource || isChrome || isData || isJavascript)
        {
            return PR_FALSE;
        }
    }
    return PR_TRUE;
}

/* Get information from the Database and store in the GtkWebHistoryItem 
 * Set the num_items parameter with the number of item in the history file
 */
nsresult GetEntries(GtkWebHistoryItem *GtkHI, gint &num_items)
{
    nsresult rv = NS_OK;

    num_items = 0;
    nsAutoString url;
    nsXPIDLCString hostname;
    nsAutoString title;
    PRTime lastVisitDate;
    const char *titlechar;

    mdb_err err;

    nsCOMPtr<nsIMdbTableRowCursor> cursor;
    nsCOMPtr<nsIMdbRow> row;

    NS_PRECONDITION(mTable != nsnull, "not initialized");
   
    err = mTable->GetTableRowCursor(mEnv, -1, getter_AddRefs(cursor));
    if (err != 0) return NS_ERROR_FAILURE;

    int marker;
    err = mTable->StartBatchChangeHint(mEnv, &marker);
    NS_ASSERTION(err == 0, "unable to start batch");
    if (err != 0) return NS_ERROR_FAILURE;

    nsCOMPtr<nsIIOService> ioService = do_GetService(NS_IOSERVICE_CONTRACTID);
    if (!ioService) return NS_ERROR_FAILURE;

    mdb_pos pos;
    do {
        err = cursor->NextRow(mEnv, getter_AddRefs(row), &pos);
        if (err != 0) return NS_ERROR_FAILURE;

        if (row)
        {
            rv = GetRowValue(row, kToken_HostnameColumn, hostname);
            rv = GetRowValue(row, kToken_URLColumn, url);
            rv = GetRowValue(row, kToken_NameColumn, title);
            rv = GetRowValue(row, kToken_LastVisitDateColumn, &lastVisitDate);
            
            NS_ConvertUCS2toUTF8 utf8ValueTitle(title.get());
            GtkHI[num_items].url = g_strdup((gchar *)url.get());
            GtkHI[num_items].title = g_strdup((gchar *)utf8ValueTitle.get());
            PRInt64 temp;
            LL_DIV(temp, lastVisitDate, PR_USEC_PER_SEC);
            LL_L2I(GtkHI[num_items].accessed, temp);

            if (IsValidEntry(g_strdup((gchar *)url.get())))
            {
                if (!url.IsEmpty() && !title.IsEmpty() && !hostname.IsEmpty())
                    num_items++;
            }
        }
    } while(row);

    // Finish the batch.
    err = mTable->EndBatchChangeHint(mEnv, &marker);
    NS_ASSERTION(err == 0, "error ending batch");

    return rv;
}

nsresult CreateTokens()
{
    mdb_err err;

    NS_PRECONDITION(mStore != nsnull, "not initialized");
    if (! mStore) return NS_ERROR_NOT_INITIALIZED;

    err = mStore->StringToToken(mEnv, "ns:history:db:row:scope:history:all", &kToken_HistoryRowScope);
    if (err != 0) return NS_ERROR_FAILURE;
  
    err = mStore->StringToToken(mEnv, "ns:history:db:table:kind:history", &kToken_HistoryKind);
    if (err != 0) return NS_ERROR_FAILURE;
  
    err = mStore->StringToToken(mEnv, "URL", &kToken_URLColumn);
    if (err != 0) return NS_ERROR_FAILURE;

    err = mStore->StringToToken(mEnv, "Referrer", &kToken_ReferrerColumn);
    if (err != 0) return NS_ERROR_FAILURE;

    err = mStore->StringToToken(mEnv, "LastVisitDate", &kToken_LastVisitDateColumn);
    if (err != 0) return NS_ERROR_FAILURE;

    err = mStore->StringToToken(mEnv, "FirstVisitDate", &kToken_FirstVisitDateColumn);
    if (err != 0) return NS_ERROR_FAILURE;

    err = mStore->StringToToken(mEnv, "VisitCount", &kToken_VisitCountColumn);
    if (err != 0) return NS_ERROR_FAILURE;

    err = mStore->StringToToken(mEnv, "Name", &kToken_NameColumn);
    if (err != 0) return NS_ERROR_FAILURE;

    err = mStore->StringToToken(mEnv, "Hostname", &kToken_HostnameColumn);
    if (err != 0) return NS_ERROR_FAILURE;
  
    err = mStore->StringToToken(mEnv, "Hidden", &kToken_HiddenColumn);
    if (err != 0) return NS_ERROR_FAILURE;

    err = mStore->StringToToken(mEnv, "Typed", &kToken_TypedColumn);
    if (err != 0) return NS_ERROR_FAILURE;

    err = mStore->StringToToken(mEnv, "GeckoFlags", &kToken_GeckoFlagsColumn);
    if (err != 0) return NS_ERROR_FAILURE;

    // meta-data tokens
    err = mStore->StringToToken(mEnv, "LastPageVisited", &kToken_LastPageVisited);

    err = mStore->StringToToken(mEnv, "ByteOrder", &kToken_ByteOrder);

    return NS_OK;
}

/* open an existing database file then get its entries */
nsresult OpenExistingFile(nsIMdbFactory *factory, const char *filePath)
{
    mdb_err err;
    nsresult rv;
    mdb_bool canopen = 0;
    mdbYarn outfmt = { nsnull, 0, 0, 0, 0, nsnull };

    nsIMdbHeap* dbHeap = 0;
    mdb_bool dbFrozen = mdbBool_kFalse; // not readonly, we want modifiable
    nsCOMPtr<nsIMdbFile> oldFile; // ensures file is released
    do
    {
        err = factory->OpenOldFile(mEnv, dbHeap, filePath,
                                    dbFrozen, getter_AddRefs(oldFile));
        copyTempFile();
    } while(!oldFile);

    // don't assert, the file might just not be there
    if ((err !=0) || !oldFile) return NS_ERROR_FAILURE;

    err = factory->CanOpenFilePort(mEnv, oldFile, // the file to investigate
                                 &canopen, &outfmt);

    // XXX possible that format out of date, in which case we should
    // just re-write the file.
    if ((err !=0) || !canopen) return NS_ERROR_FAILURE;

    nsIMdbThumb* thumb = nsnull;
    mdbOpenPolicy policy = { { 0, 0 }, 0, 0 };

    err = factory->OpenFileStore(mEnv, dbHeap, oldFile, &policy, &thumb);
    if ((err !=0) || !thumb) return NS_ERROR_FAILURE;

    mdb_count total;
    mdb_count current;
    mdb_bool done;
    mdb_bool broken;

    do
    {
        err = thumb->DoMore(mEnv, &total, &current, &done, &broken);
    } while ((err == 0) && !broken && !done);

    if ((err == 0) && done)
    {
        err = factory->ThumbToOpenStore(mEnv, thumb, &mStore);
    }

    NS_IF_RELEASE(thumb);

    if (err != 0) return NS_ERROR_FAILURE;

    rv = CreateTokens();
    NS_ENSURE_SUCCESS(rv, rv);

    mdbOid oid = { kToken_HistoryRowScope, 1 };
    err = mStore->GetTable(mEnv, &oid, &mTable);
    NS_ENSURE_TRUE(err == 0, NS_ERROR_FAILURE);
    if (!mTable)
    {
        NS_WARNING("Your history file is somehow corrupt.. deleting it.");
        return NS_ERROR_FAILURE;
    }

    err = mTable->GetMetaRow(mEnv, &oid, nsnull, getter_AddRefs(mMetaRow));
    if (err != 0)
    {
        NS_WARNING("Could not get meta row\n");
        return NS_ERROR_FAILURE;
    }
    return NS_OK;
}

nsresult GetByteOrder(char **_retval)
{ 
    NS_ENSURE_SUCCESS(OpenDB(), NS_ERROR_FAILURE);
    NS_ENSURE_ARG_POINTER(_retval);
    NS_ENSURE_STATE(mMetaRow);

    nsCAutoString byteOrder;
    mdb_err err = GetRowValue(mMetaRow, kToken_ByteOrder, byteOrder);
    NS_ENSURE_TRUE(err == 0, NS_ERROR_FAILURE);

    *_retval = ToNewCString(byteOrder);
    NS_ENSURE_TRUE(*_retval, NS_ERROR_OUT_OF_MEMORY);

    return NS_OK;
}

nsresult SaveByteOrder(const char *aByteOrder)
{
    if (PL_strcmp(aByteOrder, "BE") != 0 && PL_strcmp(aByteOrder, "LE") != 0)
    {
        NS_WARNING("Invalid byte order argument.");
        return NS_ERROR_INVALID_ARG;
    }
    NS_ENSURE_STATE(mMetaRow);

    mdb_err err = SetRowValue(mMetaRow, kToken_ByteOrder, aByteOrder);
    NS_ENSURE_TRUE(err == 0, NS_ERROR_FAILURE);

    return NS_OK;
}

nsresult InitByteOrder(PRBool aForce)
{
#ifdef IS_LITTLE_ENDIAN
    NS_NAMED_LITERAL_CSTRING(machine_byte_order, "LE");
#endif
#ifdef IS_BIG_ENDIAN
    NS_NAMED_LITERAL_CSTRING(machine_byte_order, "BE");
#endif
    nsXPIDLCString file_byte_order;
    nsresult rv = NS_OK;

    if (!aForce)
        rv = GetByteOrder(getter_Copies(file_byte_order));
    if (aForce || NS_FAILED(rv) ||
        !(file_byte_order.EqualsLiteral("BE") ||
        file_byte_order.EqualsLiteral("LE")))
    {
        // Byte order is not yet set, or needs to be reset; initialize it.
        mReverseByteOrder = PR_FALSE;
        rv = SaveByteOrder(machine_byte_order.get());
        if (NS_FAILED(rv)) return rv;
    }
    else
        mReverseByteOrder = !file_byte_order.Equals(machine_byte_order);

    return NS_OK;
}

nsresult OpenDB()
{
    static nsIMdbFactory* gMdbFactory;
    nsresult rv;

    nsCAutoString tempFileName;
    tempFileName.Assign(G_MOZILLA_TEMP_HISTORY_FILE_PATH);
  
    nsCOMPtr <nsILocalFile> historyFile(do_CreateInstance(NS_LOCAL_FILE_CONTRACTID, &rv));
    rv = historyFile->InitWithNativePath(tempFileName);
    NS_ENSURE_SUCCESS(rv, rv);

    static NS_DEFINE_CID(kMorkCID, NS_MORK_CID);
    nsCOMPtr<nsIMdbFactoryFactory> factoryfactory =
                                    do_CreateInstance(kMorkCID, &rv);
    NS_ENSURE_SUCCESS(rv, rv);

    rv = factoryfactory->GetMdbFactory(&gMdbFactory);
    NS_ENSURE_SUCCESS(rv, rv);

    mdb_err err;
    err = gMdbFactory->MakeEnv(nsnull, &mEnv);
    mEnv->SetAutoClear(PR_TRUE);
    NS_ASSERTION((err == 0), "unable to create mdb env");
    if (err != 0) return NS_ERROR_FAILURE;

    // MDB requires native file paths
    nsCAutoString filePath;
    rv = historyFile->GetNativePath(filePath);
    NS_ENSURE_SUCCESS(rv, rv);

    rv = OpenExistingFile(gMdbFactory, filePath.get());
    if (NS_FAILED(rv)) return NS_ERROR_FAILURE;

    // See if we need to byte-swap.
    InitByteOrder(PR_FALSE);

    return NS_OK;
}

nsresult copyTempFile()
{
    nsresult rv;

    nsCOMPtr <nsIFile> mozillaHistoryFile;
    rv = NS_GetSpecialDirectory(NS_APP_HISTORY_50_FILE, getter_AddRefs(mozillaHistoryFile));
    NS_ENSURE_SUCCESS(rv, rv);

    nsCOMPtr <nsIFile> tempHistoryDir;
    rv = NS_GetSpecialDirectory(NS_APP_PREFS_50_DIR, getter_AddRefs(tempHistoryDir));
    NS_ENSURE_SUCCESS(rv, rv);

    nsAutoString tempFileName;
    tempFileName.AssignLiteral(G_MOZILLA_TEMP_HISTORY_FILE_NAME);
    rv = mozillaHistoryFile->CopyTo(tempHistoryDir, tempFileName);
    NS_ENSURE_SUCCESS(rv, rv);
}

nsresult removeTempFile()
{
    nsresult rv;

    nsCAutoString tempFileName;
    tempFileName.Assign(G_MOZILLA_TEMP_HISTORY_FILE_PATH);

    nsCOMPtr <nsILocalFile> mozillaHistoryFile(do_CreateInstance(NS_LOCAL_FILE_CONTRACTID, &rv));
    rv = mozillaHistoryFile->InitWithNativePath(tempFileName);
    NS_ENSURE_SUCCESS(rv, rv);

    PRBool recursive = PR_TRUE;
    rv = mozillaHistoryFile->Remove(recursive);
    NS_ENSURE_SUCCESS(rv, rv);
}

nsresult CloseDB()
{
    if (!mStore)
        return NS_OK;

    mdb_err err;

    // order is important here - logically smallest objects first
    mMetaRow = nsnull;
  
    if (mTable)
        mTable->Release();

    mStore->Release();

    if (mEnv)
        mEnv->Release();

    mTable = nsnull; mEnv = nsnull; mStore = nsnull;

    return NS_OK;
}

void gtk_mozilla_web_get_history_list (GtkWebHistoryItem *GtkHI, gint &num_items)
{
  nsresult rv;
  
  nsCOMPtr <nsIFile> mozillaProfileDir;
  rv = NS_GetSpecialDirectory(NS_APP_PREFS_50_DIR, getter_AddRefs(mozillaProfileDir));

  nsCAutoString filePath;
  rv = mozillaProfileDir->GetNativePath(filePath);
  
  G_MOZILLA_TEMP_HISTORY_FILE_PATH = g_strconcat(filePath.get(),"/" , G_MOZILLA_TEMP_HISTORY_FILE_NAME, NULL);
  
    copyTempFile();
    OpenDB();
    GetEntries(GtkHI, num_items);
    CloseDB();
    removeTempFile();
}

void gtk_mozilla_web_remove_from_global_history(GObject *obj, GtkWebHistoryItem *GtkHI)
{
    nsresult rv;
    PRBool theRetVal = PR_TRUE;
    nsCOMPtr<nsIURI> uri;
    nsCOMPtr<nsIIOService> ios = do_GetIOService();
    if (ios)
    {		
        ios->NewURI(nsDependentCString(GtkHI->url), nsnull, nsnull,
                    getter_AddRefs(uri));
        if(!uri) return;
    }
    else return;
						
    nsCOMPtr<nsIGlobalHistory2> globalHistory(do_GetService("@mozilla.org/browser/global-history;2"));
    if (!globalHistory)
    {
        g_print("Could not get the global history object.");
        return;
    }

    nsCOMPtr<nsIBrowserHistory> myHistory = do_QueryInterface(globalHistory, &rv);
    if(NS_FAILED(rv)) {
        g_print("Could not get the history object.");
        return;
    }

    if (theRetVal)
    {
        rv = myHistory->RemovePage(uri);
    }
    else
    {
        g_print("The URL wasn't in the GH file.\r\n");
	}
}

void gtk_mozilla_web_delete_history_list(GObject *self)
{

}
